#!/usr/bin/env python

# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import multiprocessing as mp
import os
import logging
from acrilog.lib.sshlogger_pipe_handler import SSHLoggerPipeHandler


module_logger = logging.getLogger(__name__)


def main(*args, **kwargs):
    client = SSHLoggerPipeHandler(**kwargs)
    client.service_loop()


def cmdargs():
    import argparse

    filename = os.path.basename(__file__)
    progname = filename.rpartition('.')[0]

    parser = argparse.ArgumentParser(
        description="%s runs SSH logging Port Agent" % progname)
    parser.add_argument('--handler-id', type=str, dest="handler_id",
                        help="""Logger name.""")
    parser.add_argument('--log-info', type=str, dest='log_info',
                        help="""MpLogger info to using remote client.""")
    args = parser.parse_args()

    return args


if __name__ == '__main__':
    mp.freeze_support()
    mp.set_start_method('spawn')

    args = cmdargs()
    main(**vars(args))



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import pickle
import os
import logging
import socketserver
import struct
import multiprocessing as mp
import threading as th
from acrilog.lib.baselogger import BaseLogger, create_stream_handler
from acrilib import logging_record_add_host, get_free_port  # LoggerAddHostFilter
from acrilib import HierarchicalTimedSizedRotatingHandler


# TODO: get USE_QUEUE = True working without warnings at the end.
USE_QUEUE = False


class AcrilogError(Exception):
    pass


class LoggerQueueReceiver(object):

    ABORT = 'ABORT'

    def __init__(self, name=None, logging_level=None, formatter=None,
                 level_formats=None, datefmt=None, console=False,
                 args=(), kwargs={}):
        self.logger_queue = mp.Queue()
        self.name = name
        self.logging_level = logging_level
        self.formatter = formatter
        self.level_formats = level_formats
        self.datefmt = datefmt
        self.console = console
        self.args = args
        self.kwargs = kwargs

    def receiver(self,):
        logger = logging.getLogger(name=self.name)
        logger.setLevel(self.logging_level)
        handler = HierarchicalTimedSizedRotatingHandler(
            *self.args, formatter=self.formatter, **self.kwargs)
        logger.addHandler(handler)

        if self.console:
            handlers = create_stream_handler(
                logging_level=self.logging_level,
                level_formats=self.level_formats, datefmt=self.datefmt)
            for handler in handlers:
                logger.addHandler(handler)

        while True:
            record = self.logger_queue.get()
            if record:
                if record != LoggerQueueReceiver.ABORT:
                    local_logger(record, self.name)
                else:
                    self.logger_queue.close()
                    break

    def start(self):
        self.process = mp.Process(name='LoggerQueueReceiver',
                                  target=self.receiver, daemon=True)
        self.process.start()

    def stop(self):
        self.logger_queue.put(LoggerQueueReceiver.ABORT)
        self.process.join()


def local_logger(record, name=None):
    # if a name is specified, we use the named logger rather than the one
    # implied by the record.
    if name is None:
        name = record.name

    logger = logging.getLogger(name)
    # N.B. EVERY record gets logged. This is because Logger.handle
    # is normally called AFTER logger-level filtering. If you want
    # to do filtering, do it at the client end to save wasting
    # cycles and network bandwidth!
    logger.handle(record)


def get_log_record_tcp_request_handler(logger_queue=None, name=None):
    class LogRecordTCPRequestHandler(socketserver.BaseRequestHandler):
        """Handler for a streaming logging request.

        This basically logs the record using whatever logging policy is
        configured locally.
        """

        def handle(self):
            """Handles multiple requests - each expected to be a 4-byte length,
            followed by the LogRecord in pickle format. Logs the record
            according to whatever policy is configured locally.
            """
            while True:
                chunk = self.request.recv(4)
                if len(chunk) < 4:
                    break
                slen = struct.unpack('>L', chunk)[0]
                chunk = self.request.recv(slen)
                while len(chunk) < slen:
                    chunk = chunk + self.request.recv(slen - len(chunk))
                obj = self.unPickle(chunk)
                record = logging.makeLogRecord(obj)
                self.handleLogRecord(record)

        def unPickle(self, data):
            return pickle.loads(data)

        def handleLogRecord(self, record):
            if logger_queue is None:
                # set name to None so record.name will be used
                local_logger(record, name=None)
            else:
                logger_queue.put(record)

    return LogRecordTCPRequestHandler


class KeyFilter(logging.Filter):
    """
    """
    def __init__(self, name, separator='.', key='name', *args, **kwargs):
        super(KeyFilter, self).__init__(*args, **kwargs)
        self.name = name
        self.separator = separator
        self.key = key
        self.__parts = name.split(separator)

    def filter(self, record):
        name = getattr(record, self.key)
        act = name.startswith(self.name)
        return act


def start_sshlogger(name=None, host=None, port=None, handlers=[],
                    logging_level=None, formatter=None, level_formats=None,
                    datefmt=None, console=False, started=None, abort=None,
                    finished=None, args=(), kwargs={},):
    ''' starts logger for multiprocessing using queue.

    Returns:
         logger: set with correct socket handler
    '''
    # create console handler and set level to info

    logger = logging.getLogger(name=name)
    logger.setLevel(logging_level)
    logging_record_add_host()
    # logger.addFilter(LoggerAddHostFilter())

    if USE_QUEUE:
        logger_queue_receiver = LoggerQueueReceiver(
            name=name, logging_level=logging_level, formatter=formatter,
            level_formats=level_formats, datefmt=datefmt, console=console,
            args=args, kwargs=kwargs)
        logger_queue_receiver.start()
        logger_queue = logger_queue_receiver.logger_queue
    else:
        logger_queue = None
        handlers += [
            HierarchicalTimedSizedRotatingHandler(*args,
                                                  formatter=formatter,
                                                  **kwargs)]
        if console:
            handlers += create_stream_handler(logging_level=logging_level,
                                              level_formats=level_formats,
                                              datefmt=datefmt)
        for handler in handlers:
            logger.addHandler(handler)

    handler = get_log_record_tcp_request_handler(logger_queue=logger_queue,
                                                 name=name)
    tcpserver = socketserver.ThreadingTCPServer((host, port), handler)
    tcpserver.allow_reuse_address = True
    tcpserverproc = th.Thread(name='ThreadingTCPServer',
                              target=tcpserver.serve_forever, daemon=True)
    tcpserverproc.start()
    # notify caller, process started
    started.set()
    # wait for abort notification
    abort.wait()
    # abort received, shoutdown TCP server
    tcpserver.shutdown()
    tcpserver.server_close()
    tcpserverproc.join()
    if USE_QUEUE:
        logger_queue_receiver.stop()
    finished.set()


class SSHLogger(BaseLogger):
    def __init__(self, name=None, host='localhost', port=None,
                 logging_level=logging.INFO, handlers=[], *args, **kwargs):
        super(SSHLogger, self).__init__(*args, name=name,
                                        logging_level=logging_level, **kwargs)

        self.host = host
        self.logger_initialized = False
        self.logging_level = logging_level
        self.handlers = handlers
        self.args = args
        # self.kwargs = kwargs

        if port is None:
            try:
                port = get_free_port()
            except Exception as e:
                raise AcrilogError('Failed to get free port.') from e
            if port is None:
                raise AcrilogError('Failed to get free port, got None.')
        #self.port = logging.handlers.DEFAULT_TCP_LOGGING_PORT
        self.port = port

    def logger_info(self):
        info = super(SSHLogger, self).logger_info()
        info.update({
                'host': self.host,
                'port': self.port,
               })
        return info

    @classmethod
    def get_logger(cls, logger_info, name=None):
        # create the logger to use.
        # logger = BaseLogger.get_logger(logger_info, name)

        try:
            # name = name if name is not None else logger_info['name']
            # We don't really need host, since it is always localhost
            # host = logger_info['host']
            host = 'localhost'
            port = logger_info['port']
        except Exception as e:
            msg = "Failed to get info from logger_info: {}"
            raise AcrilogError(msg.format(repr(logger_info))) from e

        logger = BaseLogger.get_logger(logger_info, name)

        # check logger has already proper handlers or not
        already_set = False
        for handler in logger.handlers:
            if isinstance(handler, logging.handlers.SocketHandler):
                already_set = already_set or (handler.port == port and
                                              handler.host == host)

        if not already_set:
            socketHandler = logging.handlers.SocketHandler(host, port)
            # socket handler sends the event as an unformatted pickle
            logger.addHandler(socketHandler)
            logging_record_add_host()
            # logger.addFilter(LoggerAddHostFilter())
        return logger

    def start(self, name=None):
        ''' starts socket based logger server.

        Args:
            name: to associate with local logger (as oppose to server logger).

        Returns:
            local logger object.
        '''
        self.started = mp.Event()
        self.abort = mp.Event()
        self.finished = mp.Event()

        start_nwlogger_kwargs = {
            'name': self.name,
            'host': self.host,
            'port': self.port,
            'handlers': self.handlers,
            'logging_level': self.logging_level,
            'formatter': self.record_formatter,
            'level_formats': self.level_formats,
            'datefmt': self.datefmt,
            'console': self.console,
            'started': self.started,
            'abort': self.abort,
            'finished': self.finished,
            'args': self.handler_args,
            'kwargs': self.handler_kwargs,
        }

        self.tcpserver = mp.Process(name='NwLogger', target=start_sshlogger,
                                    kwargs=start_nwlogger_kwargs, daemon=False)
        self.tcpserver.start()
        self.started.wait()

        logger = SSHLogger.get_logger(self.logger_info(), name=name)
        return logger

    def stop(self,):
        if self.abort:
            self.abort.set()
            self.finished.wait()
            self.tcpserver.terminate()
            self.tcpserver.join()



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import logging
import multiprocessing as mp
from copy import copy
from acrilib import get_file_handler, get_hostname
from acrilib import LevelBasedFormatter
from copy import deepcopy
import sys


def create_stream_handler(logging_level=logging.INFO, level_formats={},
                          datefmt=None, encoding='utf8'):
    handlers = list()

    stdout = sys.stdout
    # stderr = sys.stderr

    stdout_handler = logging.StreamHandler(stream=stdout)
    # stdout_handler.setLevel(logging_level)
    formatter = LevelBasedFormatter(level_formats=level_formats,
                                    datefmt=datefmt)
    stdout_handler.setFormatter(formatter)
    handlers.append(stdout_handler)

    # stderr_handler = logging.StreamHandler(stream=stderr)

    return handlers


class BaseLogger(object):
    ''' Builds Multiprocessing logger such all process
        hare the same logging mechanism.
    '''

    kwargs_defaults = {
        'logdir': '/tmp',
        'key': 'name',
        'file_mode': 'a',
        'file_prefix': '',
        'file_suffix': '',
        'maxBytes': 0,
        'backupCount': 0,
        'encoding': 'ascii',
        'delay': False,
        'when': 'h',
        'interval': 1,
        'utc': False,
        'atTime': None,
       }

    logger_info_defaults = {
        'level_formats': {
                logging.DEBUG: "[ %(asctime)-15s ][ %(levelname)-7s ][ %(host)s ][ %(processName)-11s ][ %(message)s ][ %(module)s.%(funcName)s(%(lineno)d) ]",
                'default': "[ %(asctime)-15s ][ %(levelname)-7s ][ %(host)s ][ %(processName)-11s ][ %(message)s ]",
                },
         'datefmt': '%Y-%m-%d,%H:%M:%S.%f',
         'handler_kwargs': kwargs_defaults,
            }

    def __init__(self, name=None, logging_level=logging.INFO, level_formats={}, datefmt=None, console=False, handlers=[], *args, **kwargs):
        '''
        Args:
            name: base name to use for file logs.
            logdir: folder to which log files will be written; if not provided, log files will not be created
            logging_level: level from which logging will be done 
            level_formats: mapping of logging levels to formats to use for constructing message
            datefmt: date format to use
            process_key: list of record names that would be used to create files
            console_name: when set, records assigned to process_key handler will also routed to global handlers.
            #logging_root: defaults to name if not provided
            encoding: used in defining file handlers; default 'ascii'
            handlers: list of global handlers
            kwargs: pass-through to handler defining its policy
                key='name',
                file_mode='a',
                file_prefix='',
                file_suffix='',
                maxBytes=0,
                backupCount=0,
                delay=False,
                when='h',
                interval=1,
                utc=False,
                atTime=None

        '''

        self.logging_level = logging_level
        self.level_formats = level_formats if level_formats else BaseLogger.logger_info_defaults['level_formats']
        self.datefmt = datefmt if datefmt else BaseLogger.logger_info_defaults['datefmt']
        self.record_formatter = LevelBasedFormatter(level_formats=self.level_formats, datefmt=self.datefmt)
        self.logger_initialized = False
        self.handlers = handlers
        # self.process_key = process_key
        # self.consolidate = consolidate
        self.console = console
        self.name = name if name is not None else 'logger'
        self.handler_kwargs = copy(BaseLogger.kwargs_defaults)
        self.handler_kwargs.update(kwargs)
        self.handler_args = args
        # self.encoding=encoding
        # self.file_mode=file_mode
        # self.local_log = local_log

    def global_file_handlers(self,):
        # if not process_key: process_key=self.name
        # handlers = get_file_handler(logging_level=self.logging_level, formatter=self.record_formatter, **self.kwargs)
        handler = get_file_handler(formatter=self.record_formatter, **self.kwargs)
        self.global_filename = handler.filename
        return [handler]
        # for handler in handlers:
        #     self.queue_listener.addHandler(handler)  

    @classmethod
    def add_file_handlers(cls, logger,  record_formatter, **kwargs):
        '''
        Args:
            kwargs:
                file_mode='a', 
                maxBytes=0, 
                backupCount=0, 
                encoding='ascii', 
                delay=False, 
                when='h', 
                interval=1, 
                utc=False, 
                atTime=None
        '''
        # if not process_key: process_key=name
        # global_handlers = get_file_handler(logging_level=logging_level, formatter=record_formatter, **kwargs)
        global_handlers = [get_file_handler(formatter=record_formatter, **kwargs)]

        for handler in global_handlers:
            logger.addHandler(handler) 

    @classmethod
    def base_info(cls, log_info):
        keys = [
            'name',
            'logging_level',
            'level_formats',
            'datefmt',
            'handler_kwargs',
            'server_host',
            ]
        result = dict([(k,log_info[k]) for k in keys])
        return result

    def logger_info(self):
        hostname = get_hostname()
        info = {'name': self.name,
                # 'process_key': self.process_key,
                # 'logdir': self.logdir, 
                'logging_level': self.logging_level,
                'level_formats': self.level_formats,
                'datefmt': self.datefmt,
                'handler_kwargs': self.handler_kwargs,
                'server_host': hostname,
               }
        return info

    @classmethod
    def get_logger(cls, logger_info, name=None, local=False):
        # create the logger to use.
        name = name if name is not None else logger_info['name']
        logger = logging.getLogger(name)
        logger.propagate = False

        # TODO: CHANGE TO USE MergedChainedDict
        defaults = deepcopy(BaseLogger.logger_info_defaults)
        defaults.update(logger_info)
        logger_info = defaults
        # The only handler desired is the SubProcessLogHandler.  If any others
        # exist, remove them. In this case, on Unix and Linux the StreamHandler
        # will be inherited.

        this_host = get_hostname() #(full=True)
        # server_host = logger_info.get('server_host', 'localhost')
        # server may already started logger
        # if logger_info['server_host'] == server_host: return logger

        # add the handler only if processing locally and this host is not server host.
        # if server_host == 'localhost' or server_host != this_host:
        if local:
            level_formats = logger_info['level_formats']
            datefmt = logger_info['datefmt']
            cls.add_file_handlers(name=name,
                                  # process_key=logger_info['process_key'], 
                                  logger=logger,
                                  # logdir=logger_info['logdir'], 
                                  # logging_level=logging_level,
                                  record_formatter=LevelBasedFormatter(level_formats=level_formats, datefmt=datefmt),
                                  **logger_info['handler_kwargs'],
                                  )

        # On Windows, the level will not be inherited.  Also, we could just
        # set the level to log everything here and filter it in the main
        # process handlers.  For now, just set it from the global default.
        logging_level = logger_info['logging_level']
        logger.setLevel(logging_level)     
        return logger

    
if __name__ == '__main__':
    mp.freeze_support()
    mp.set_start_method('spawn')



# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

import logging
import signal
from logging.handlers import QueueListener, QueueHandler
import multiprocessing as mp
from acrilog.lib.baselogger import BaseLogger, create_stream_handler, get_file_handler
# from acrilib import LoggerAddHostFilter
from acrilib import logging_record_add_host, HierarchicalTimedSizedRotatingHandler
# import threading as th


class _QueueListener(QueueListener):
    def __init__(self, started, *args, **kwargs):
        super(_QueueListener, self).__init__(*args, **kwargs)
        started.set()

    def dequeue(self, block,):
        ''' adding capture to EOF
        '''
        try:
            item = self.queue.get(block)
        except EOFError:
            item = None
        return item


class MpQueueListener(QueueListener):
    def __init__(self, queue, name=None, logging_level=logging.INFO, logdir=None, formatter=None, process_key=['processName'], global_handlers=[], **kwargs):
        super(MpQueueListener, self).__init__(queue, *global_handlers)
        """ Initialize an instance with the specified queue and
        handlers.

        Args:
            handlers: list of handlers to apply
            process_key: list of keys by which to bind handler to records.
                handlers that don't have any key are classified as global handlers.
                if record doens't have any matching key, global handlers will be used.
                if records match, only matching handlers will be used. 
        """
        self.process_key = process_key
        self.logdir = logdir
        self.formatter = formatter
        self.name = name
        self.kwargs = kwargs

        key_handlers = dict([(p, dict()) for p in process_key])

        self.key_handlers = key_handlers
        self.global_handlers = global_handlers
        self.console_handlers = list()
        self.logging_level = logging_level

    def handle(self, record):
        """ Override handle a record.
        This just loops through the handlers offering them the record
        to handle.
        Args:
            record: The record to handle.
        """

        # Find handlers that match process keys
        handlers = list()
        record_name = record.__dict__.get('name', None)
        for process_key in self.process_key:
            record_key = record.__dict__.get(process_key, None)
            # print('record_key[process_key]: %s[%s]' %(record_key, process_key))
            # print('record_key[processName]: %s' %(repr(record.__dict__)))
            if record_key: 
                process_handlers = self.key_handlers[process_key]
                key_handlers = process_handlers.get(record_key, [])

                # avoid getting dedicated handler in special case when in consolidated mode and record with 
                # name equal to the global one (QueueListiner name)
                need_handler = len(key_handlers) ==0 and (record_key != self.name or len(self.global_handlers) ==0)
                if need_handler:
                    name = record_name
                    # file_prefix=self.kwargs.get('file_prefix')
                    # if file_prefix is None: file_prefix=name
                    # print('file_prefix, record_key, record_name:', file_prefix, record_key, record_name)
                    if record_name != record_key:
                        name = "%s.%s" % (name, record_key)
                    key_handlers = get_file_handler(logging_level=self.logging_level, logdir=self.logdir, process_key=name, formatter=self.formatter, **self.kwargs)
                    process_handlers[record_key] = key_handlers
                handlers.extend(key_handlers)

        if len(self.global_handlers) > 0:
            handlers.extend(self.global_handlers)

        if len(self.console_handlers) > 0:
            handlers.extend(self.console_handlers)

        record = self.prepare(record)

        for handler in list(set(handlers)):
            if record.levelno >= handler.level: # This check is not in the parent class
                handler.handle(record)

    def addConsoleHandler(self, handler):
        self.console_handlers.append(handler)

    def addHandler(self, handler):
        """
        Add the specified handler to this logger.

        handler is expected to have process_key attribute.
        process_key attribute is expected to be a list of records attribute names that handler would bind to.
        if handler does not have process_key attribute or it is empty, handler will be associated with 
        """
        key_bind=False
        if hasattr(handler, 'process_key'):
            handler_key=handler.process_key
            for key in list(set(self.process_key) & set(handler_key)):
                exist_handler=self.key_handlers.get(key, list())
                self.key_handlers[key]=exist_handler
                exist_handler.append(handler)
                key_bind=True
        if not key_bind:
            self.global_handlers.append(handler)

    def removeHandler(self, hdlr):
        """
        Remove the specified handler from this logger.
        """
        if hdlr in self.handlers:
            hdlr.close()
            self.handlers.remove(hdlr)


def _start(name=None, loggerq=None, handlers=[], logging_level=None,
           formatter=None, level_formats=None, datefmt=None,
           console=False, started=None, abort=None, finished=None,
           verbose=False,
           args=(), kwargs={},):
    logger = logging.getLogger(name)
    logger.setLevel(logging_level)
    logging_record_add_host()
    # logger.addFilter(LoggerAddHostFilter())

    handlers += [HierarchicalTimedSizedRotatingHandler(
        *args, formatter=formatter, **kwargs)]

    if console:
        console_handlers = \
            create_stream_handler(logging_level=logging_level,
                                  level_formats=level_formats, datefmt=datefmt)
        handlers.extend(console_handlers)

    for handler in handlers:
        logger.addHandler(handler)

    queue_listener = _QueueListener(started, loggerq, *handlers)

    # queue_listener = LogRecordQueueListener(loggerq, verbose=verbose)
    if verbose:
        print('start_mplogger: starting listener.')
    queue_listener.start()
    # started.set()
    if verbose:
        print('start_mplogger: listener started.')
    # return queue_listener

    def exit_gracefully(signo, stack_frame, *args, **kwargs):
        queue_listener.stop()
        finished.set()

    # set exits
    signal.signal(signal.SIGHUP, exit_gracefully)
    signal.signal(signal.SIGINT, exit_gracefully)
    signal.signal(signal.SIGTERM, exit_gracefully)

    # waiting on abort prevents creates a deadlock
    # looging shutdown therefore will not be called automatically.
    # therefore, shutdown needs to be explicit.
    abort.wait()
    queue_listener.stop()
    if verbose:
        print('start_mplogger: setting finished.')
    finished.set()


class MpLogger(BaseLogger):
    ''' Builds Multiprocessing logger such all process share the same
    logging mechanism
    '''

    def __init__(self, name=None, logging_level=logging.INFO, handlers=[],
                 verbose=False, *args, **kwargs):
        '''Initiates MpLogger service

        Args:
            name: base name to use for file logs.
            logdir: folder to which log files will be written; if not provided,
                log files will not be created logging_level: level from which
                logging will be done
            level_formats: mapping of logging levels to formats to use for
                constructing message
            datefmt: date format to use
            process_key: list of record names that would be used to create
                files
            console_name: when set, records assigned to process_key handler
                will also routed to global handlers.
            encoding: used in defining file handlers; default 'ascii'
            handlers: list of global handlers
            kwargs: pass-through to hierarchical handler defining its policy
                key='name'
                separator='.'
                consolidate=''
                file_mode='a',
                file_prefix='',
                file_suffix='',
                maxBytes=0,
                backupCount=0,
                encoding='ascii',
                delay=False,
                when='h',
                interval=1,
                utc=False,
                atTime=None

        '''
        super(MpLogger, self).__init__(*args, name=name,
                                       logging_level=logging_level, **kwargs)

        self._queue_listener = None
        self.abort = None
        self.logger_initialized = False
        self.handlers = handlers
        self.verbose = verbose

    def logger_info(self):
        info = super(MpLogger, self).logger_info()
        info.update({
                'loggerq': self.loggerq,
               })
        return info

    @classmethod
    def get_logger(cls, logger_info, name=None):
        # create the logger to use.
        if name is not None:
            logger_info['name'] = name
        logger = BaseLogger.get_logger(logger_info)
        # logger.addFilter(LoggerAddHostFilter())
        logging_record_add_host()
        loggerq = logger_info['loggerq']

        # check logger has already proper handlers or not
        already_set = False
        for handler in logger.handlers:
            if isinstance(handler, QueueHandler):
                already_set = already_set or (handler.queue == loggerq)

        if not already_set:
            queue_handler = QueueHandler(loggerq)
            logger.addHandler(queue_handler)

        return logger

    def start(self, name=None):
        ''' starts logger for multiprocessing using queue.

        Args:
            name: identify starting process to allow it log into its own
                logger.

        Returns:
            logger: set with correct Q handler.
        '''
        # create console handler and set level to info

        if self.logger_initialized:
            return

        self.logger_initialized = True

        self.loggerq = mp.Queue()

        # self._manager = manager = mp.Manager()
        self.abort = mp.Event()
        started = mp.Event()
        self.finished = mp.Event()

        start_kwargs = {
            'name': self.name,
            'loggerq': self.loggerq,
            'handlers': self.handlers,
            'logging_level': self.logging_level,
            'formatter': self.record_formatter,
            'level_formats': self.level_formats,
            'datefmt': self.datefmt,
            'console': self.console,
            'started': started,
            'abort': self.abort,
            'finished': self.finished,
            'args': self.handler_args,
            'kwargs': self.handler_kwargs,
            'verbose': self.verbose,
            }

        #self._queue_listener = _start(**start_kwargs)

        self._queue_listener = \
            mp.Process(target=_start, kwargs=start_kwargs, daemon=True)
        # self._queue_listener = th.Thread(target=start_mplogger,
        #                               kwargs=start_kwargs, daemon=False)
        self._queue_listener.start()

        started.wait()

        logger_info = self.logger_info()
        logger = MpLogger.get_logger(logger_info=logger_info, name=name)
        return logger

    def __exit__(self):
        self.__del__()

    def __del__(self):
        self.stop()

    def stop(self,):
        # self._queue_listener.join()
        # return
        if self.verbose:
            print('mplogger stopping.')
        if self._queue_listener:
            if self.verbose:
                print('mplogger stop: stopping queue listener.')
            if self.abort:
                if self.verbose:
                    print('mplogger stop: setting abort.')
                self.abort.set()
                # self._queue_listener.stop()
                if self.verbose:
                    print('mplogger stop: waiting to finish.')
                self.finished.wait()
                if self.verbose:
                    print('mplogger stop: joining process.')
                self._queue_listener.join()


if __name__ == '__main__':
    mp.freeze_support()
    mp.set_start_method('spawn')



#!/usr/bin/env python

# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import logging
from acrilog import SSHLogger
import yaml
from sshpipe import SSHPipeHandler


module_logger = logging.getLogger(__name__)


class SSHLoggerPipeHandler(SSHPipeHandler):

    def __init__(self, log_info=None, *args, **kwargs):
        global module_logger
        super(SSHLoggerPipeHandler, self).__init__(*args, **kwargs)
        try:
            log_info = yaml.load(log_info)
        except Exception as e:
            msg = "Failed to YAML.load('{}')".format(log_info)
            raise RuntimeError(msg) from e

        # TODO: why do we need to do such assignment
        #       if logger has proper handler
        module_logger = self.module_logger
        module_logger.debug('Accepted logging info:\n    {}.'.format(log_info))
        self.sshlogger = SSHLogger.get_logger(log_info)

    def handle(self, received):
        # it may be "TERM" message or alike
        module_logger.debug('Handling record:\n    {}.'
                            .format(repr(received)))
        if isinstance(received, logging.LogRecord):
            self.sshlogger.handle(received)



#!/usr/bin/env python

# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
import os
import sshpipe
import logging
from copy import deepcopy
from acrilib import logging_record_add_host  # LoggerAddHostFilter
from acrilog import MpLogger, SSHLogger
import yaml
from sshpipe import SSHPipeHandler


module_logger = logging.getLogger(__name__)


class SSHLoggerHandlerError(Exception):
    pass


class LoggingSSHPipeHandler(SSHPipeHandler):

    def __init__(self, log_info=None, *args, **kwargs):
        global module_logger
        super(LoggingSSHPipeHandler, self).__init__(*args, **kwargs)
        try:
            log_info = yaml.load(log_info)
        except Exception as e:
            msg = "Failed to YAML.load('{}')".format(log_info)
            raise Exception(msg) from e

        # TODO: why do we need to do such assignment
        #       if logger has proper handler
        module_logger = self.module_logger
        module_logger.debug('Accepted logging info:\n    {}.'.format(log_info))
        self.sshlogger = SSHLogger.get_logger(log_info)

    def handle(self, received):
        # it may be "TERM" message or alike
        module_logger.debug('Handling record:\n    {}.'
                            .format(repr(received)))
        if isinstance(received, logging.LogRecord):
            self.sshlogger.handle(received)


class SSHLoggerClientHandler(logging.Handler):
    ''' SSHPipe Logging handler to send logging records to remote logging server

    SSHLoggerClientHandler create handler object that sends
    '''

    def __init__(self, logger_info, ssh_host, verbose=False):
        ''' Initiate logger client on remote connecting to host:port

        Args:
            logger_info: result of SSHLogger.logger_info().
            ssh_host: SSH config Host to connect to.
        '''
        global module_logger
        super(SSHLoggerClientHandler, self).__init__()
        self.logger_info = logger_info
        self.sshpipe = None
        self.verbose = verbose

        mp_logger_params = deepcopy(logger_info)
        del mp_logger_params['port']
        name = mp_logger_params['name']
        handler_id = name + '_sshlogger_client_handler'
        mp_logger_params['name'] = handler_id
        # handler_id = mp_logger_params['name'] + '_sshlogger_socket_handler'
        handler_kwargs = mp_logger_params.get('handler_kwargs', dict())
        kwargs = {}
        kwargs.update(mp_logger_params)
        kwargs.update(handler_kwargs)

        self._mp_logger = MpLogger(**kwargs, verbose=self.verbose)
        self._mp_logger.start()
        mp_logger_info = self._mp_logger.logger_info()
        module_logger = MpLogger.get_logger(mp_logger_info, )

        logging_record_add_host()
        # self.addFilter(LoggerAddHostFilter())

        # there is no need to pass loggerq via ssh.
        # also, it wont work anyhow.
        # but it does need port
        del mp_logger_info['loggerq']
        mp_logger_info['port'] = logger_info['port']
        # mp_logger_info['console'] = False

        # this program has a twin in bin as executable
        command = ["{}".format(os.path.basename(__file__)), ]
        # server_host = logger_info['server_host']

        # logger_name = logger_info['name']
        kwargs = {"--handler-id": handler_id,  # logger_name,
                  # "--host": server_host, #logger_info['host'],
                  # "--port": logger_info['port'],
                  '--log-info': '"{}"'.format(yaml.dump(mp_logger_info)),
                  # "--logging-level": logger_info['logging_level'],
                  # "--server-host": logger_info['server_host'],
                  # "--logdir": logdir,
                  }
        command.extend(["{} {}".format(name, value)
                        for name, value in kwargs.items()])
        command = ' '.join(command)

        logname = logger_info['name']

        try:
            # Important: cannot pass logger from here to SSHPipe
            # If logger is passed, infinite recursion is created.
            self.sshpipe = sshpipe.SSHPipe(ssh_host, command,
                                           name=logname, logger=module_logger)
            msg = "Starting remote logger SSHPipe on host: {}, command: {}"
            module_logger.debug(msg.format(ssh_host, command))
            self.sshpipe.start()
        except Exception as e:
            module_logger.exception(e)
            module_logger.critical(
                'Agent failed to start: {}'.format(ssh_host))
            response = self.sshpipe.response()
            raise SSHLoggerHandlerError(
                ("Failed to start SSHPipe to: {}; "
                 "response: {}.").format(ssh_host, response))

        if not self.sshpipe.is_alive():
            module_logger.critical(
                'Agent process terminated unexpectedly: {}.'.format(ssh_host))
            self.sshpipe.join()
            response = self.sshpipe.response()
            raise SSHLoggerHandlerError(
                ("Failed to start SSHPipe to: {}; "
                 "response: {}.").format(ssh_host, response))

        if self.verbose:
            print("SSHLoggerClientHandler: Remote logger SSHPipe started.", self.sshpipe.is_alive())
        module_logger.debug("Remote logger SSHPipe started.")

    def emit(self, record):
        if self.verbose:
            print("SSHLoggerClientHandler: emitting record:", repr(record))
        try:
            self.sshpipe.send(record)
        except Exception as e:
            raise e
            raise SSHLoggerHandlerError(
                "Failed SSHPipe send: {}.".format(record.msg)) from e

    def __del__(self):
        if self._mp_logger:
            self.close()

    def close(self):
        if self.verbose:
            print("SSHLoggerClientHandler: closing handler.")
        if self._mp_logger:
            if self.verbose:
                print("SSHLoggerClientHandler: stopping mplogger.")
            self._mp_logger.stop()
            self._mp_logger = None
        if self.sshpipe.is_alive():
            if self.verbose:
                print("SSHLoggerClientHandler: closing sshpipe.")
            self.sshpipe.close()
        if self.verbose:
            print("SSHLoggerClientHandler: closing hanlers.")
        #super(SSHLoggerClientHandler, self).close()
